; Warning: First, seting the package where the functions will be defined 
(in-package :om)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Specific classes that have reactive properties
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;To be anotoized
;declare the type of ed box as reactive-box
;implements the elements that could be called from anoto-decode-method 

;this class represents an anoto-object and redefine menu to export to paper designer server.
;menu is accessible from the objet (right click) and from the export menu

(defclass! anoto-object () ((trigger :initarg :trigger :initform :play :accessor trigger)))

(defmethod export-menu-from-obj ((self omboxeditcall) (obj anoto-object))
  (list (call-next-method)
        (om-new-leafmenu "Export to Anoto" #'(lambda () 
                                               (cond (*anoto-print*
                                                      (export-anoto-for-object obj (name self)))
                                                     (t
                                                      (om-beep-msg "Anoto Server Not Connected")))))))

(defmethod score-export ((format (eql 'anoto)) (obj anoto-object) edition-params)
  (cond (*anoto-print*
         (export-anoto-for-object obj (name obj)))
        (t
         (om-beep-msg "Anoto Server Not Connected"))))


(defmethod* objFromObjs ((self t) (type anoto-object))
  (if (subtypep (type-of type) (type-of self)) 
      (let ((rep (clone self)))
        (change-class rep (class-of type))
        rep)
    ))


;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;REACTIVE BOX;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;
;class reactive box to use reactive paradigm dans un objet OM

(defclass reactive-box (omboxeditcall) ())

(defclass reactive-box-frame (boxEditorFrame) ())

(defmethod get-frame-class ((self reactive-box)) 'reactive-box-frame)

;used to add the objet in box list
(defmethod make-frame-from-callobj ((self reactive-box))
  (push self *anoto-box-list*)
  (call-next-method))

;used to remove the object in box list
(defmethod close-frame ((box reactive-box-frame)) 
  (call-next-method)
  (setf *anoto-box-list* (remove (object box) *anoto-box-list* :test 'equal)))

(defmethod anoto-decode-message ((self reactive-box) command args)
  (let ((funname command))
    (cond 
          ((string-equal funname "clear")
           (anoto-clear (value self))
           (update-if-editor self))
          (t (anoto-decode-message-for-object (value self) self command args)))))

(defmethod anoto-decode-message ((self t) command args) 
  nil)

(defmethod export-anoto-for-object ((self t) name) 
  (print (list self name "export method does no exist")))

(defmethod anoto-clear ((self t))
  t)
;;;;;;;;;;;;
;chord-seq-r
;;;;;;;;;;;;

(defclass! chord-seq-r (anoto-object chord-seq) ())

(defmethod get-type-of-ed-box ((self chord-seq-r)) 'reactive-box)
  
;format is (addnote pitch onset duration)
(defmethod anoto-decode-message-for-object ((self chord-seq-r) reactive-box command args)
  (let ((funname command))
    (cond ((string-equal funname "addnote")
           (anoto-addnote self (car args) (caddr args) (cadr args))
           (update-if-editor reactive-box))
          ((string-equal funname "play")
           (anoto-play self)
           (update-if-editor reactive-box))
          (t 
          (print (format nil "commande ~A not found!" command))))))

(defmethod anoto-clear ((self chord-seq-r))
  (setf (inside self) nil) 
  (setf (Qvalue self) 1000)
  (adjust-extent self)
  (QNormalize self)
  t)

(defmethod anoto-addnote ((self chord-seq-r) pitch dur onset)
  (let ((new-chord (mki 'chord :Lmidic (list pitch) :ldur (list dur)))
        new-chord-list)
    (setf (offset new-chord) onset)    
    (loop for item in (chords self) do
          (setf (offset item) (offset->ms item)))
    (setf new-chord-list (cons new-chord (chords self)))
    (setf (inside self) (sort new-chord-list '< :key 'offset)) 
    (setf (Qvalue self) 1000)
    (adjust-extent self)
    (QNormalize self)
    t))

(defmethod anoto-play ((self chord-seq-r))
  (play self))

(defmethod! play-as (cseq type)
  (let ((obj2 (if (equal type :arpege)
                  cseq
                (make-instance 'chord 
                               :lmidic (flat (lmidic cseq))
                               :lvel (flat (lvel cseq))))))
    (play obj2)))
                  


(defmethod export-formats ((self chord-seq-r))
  (append (call-next-method)
          (list '(anoto "Anoto"))))


(defmethod export-anoto-for-object ((self chord-seq-r) name)
  (anoto-osc-send (list (anoto-data-address) "add" "PChordSeq" name))
  (anoto-osc-send (list (add-anoto-data-address name) "clear"))
  (loop for midics in (LMidic self)
        for onset in (LOnset self)
        for durs in (Ldur self)
        do
        (loop   for midic in midics   
                for dur in durs
                do
                (anoto-osc-send (list (add-anoto-data-address name) "addnote" midic onset dur)))))

;;;;;;;;;;;;
;bpf-r
;;;;;;;;;;;;

(defclass! bpf-r (anoto-object bpf) ())

(defmethod get-type-of-ed-box ((self bpf)) 'reactive-box)

(defmethod anoto-decode-message-for-object ((self bpf-r) reactive-box command args)
  (let ((funname command))
    (cond ((string-equal funname "addpoint")
           (anoto-addpoint self (car args) (cadr args))
           (update-if-editor reactive-box))
          ((string-equal funname "setrange")
           (anoto-setrange self (car args) (second args) (third args) (fourth args))
           (update-if-editor reactive-box))
          (t 
          (print (format nil "commande ~A not found!" command))))))


(defmethod anoto-clear ((self bpf-r))
   (setf (point-list self) nil)
  t)

(defmethod anoto-addpoint ((self bpf-r) x_coord y_coord)
   (insert-point self (om-make-point (* x_coord (expt 10 (decimals self))) (* y_coord (expt 10 (decimals self)))))
    t) 

(defmethod anoto-setrange ((self bpf-r) x_min x_max y_min y_max)
  t)

(defmethod export-formats ((self bpf))
  (append (call-next-method)
          (list '(anoto "Anoto"))))


(defmethod export-anoto-for-object ((self bpf-r) name)
  (anoto-osc-send (list (anoto-data-address) "add" "PCurve" name))
  (anoto-osc-send (list (add-anoto-data-address name) "clear"))
  (anoto-osc-send (list (add-anoto-data-address name) "setrange" (first (get-scaled-bpf-range self)) (second (get-scaled-bpf-range self)) (third (get-scaled-bpf-range self)) (fourth (get-scaled-bpf-range self))) )
  (loop for point in (point-pairs self)
        do
        (anoto-osc-send (list (add-anoto-data-address name) "addpoint" (car point) (car (cdr point))))))

(defmethod get-scaled-bpf-range ((self bpf))
  (om/ (give-bpf-range self) (om^ 10 (float (decimals self)) )))

;;;;;;;;;;;;
;bpc-r
;;;;;;;;;;;;

;bpc first to get the editor from bpc and anoto-method from bpf-r
;might be better to create a specific class with anoto functionnalities...
(defclass! bpc-r (bpc bpf-r) ())

;need to redefine because already existing in bpc...
(defmethod* objFromObjs ((self bpc) (type bpc-r))
  (if (subtypep (type-of type) (type-of self)) 
      (let ((rep (clone self)))
        (change-class rep (class-of type))
        rep)
    ))

;;;;;;;;;;;;
;trigger
;;;;;;;;;;;;

(defclass! trigger (anoto-object named-object)
           ((message :initform "eval" :initarg :message :accessor message)
            (eval-inlet :initform nil :initarg :eval-inlet :accessor eval-inlet))
            (:documentation "anoto trigger to export and use in paper designer UI")
            (:icon 112))

(defmethod get-type-of-ed-box ((self trigger)) 'reactive-box)

;there is a bug here....
(defmethod anoto-decode-message-for-object ((self trigger) reactive-box command args)
  (let ((funname command))
    (cond ((string-equal funname "trigger")
           (omng-box-value reactive-box))
          ((string-equal funname "settrigger")
           (setf (message self) (car args))
           (update-if-editor reactive-box))
          (t 
          (print (format nil "commande ~A not found!" command))))))

;setf slot value class name;;;

(defmethod anoto-clear ((self trigger))
  t)

(defmethod export-formats ((self trigger))
  (append (call-next-method)
          (list '(anoto "Anoto"))))

(defmethod export-anoto-for-object ((self trigger) name)
  (anoto-osc-send (list (anoto-data-address) "add" "PTrigger" name))
  (anoto-osc-send (list (add-anoto-data-address name) "settrigger" (message self))))


;;;;;;;;;;;;;;;;;
;;; Export for several components
;;;;;;;;;;;;;;;;;

(defmethod export-anoto-for-object ((self note) name)
 (print (append (list (midic self) (dur self) (vel self) (tie self) ))) )


;;;;;;;;;;;;;;;;;
;;; Trigger utilities
;;;;;;;;;;;;;;;;;


(defmethod! anoto-route (trigger value)
            (equal trigger value))

(defmethod! anoto-eval-if (trigger value box)
           (omif (anoto-route trigger value) box))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Specific reactive score, override usual scores
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;
; poly et voice;;;;;
;;;;;;;;;;;;;;;;;;;;

(defmethod get-type-of-ed-box ((self poly)) 'reactive-box)
  

(defmethod export-menu ((self scoreEditor))
  (when (export-formats (object self))
    (om-make-menu "Export" 
                  (append (mapcar #'(lambda (item) 
                                      (om-new-leafmenu (cadr item) #'(lambda () 
                                                                       (score-export (car item) (object self) (editor-edit-params self)))))
                                  (export-formats (object self)))
                          (list (om-new-leafmenu "Anoto" #'(lambda () 
                                                             (score-export 'anoto (object self) self))
                                                 "f")
                                )))))


(defmethod score-export ((format (eql 'anoto)) object scoreEditor)
  (create-anoto-file scoreEditor)
  ;(create-anoto-image scoreEditor)
  )


(defmethod create-anoto-image ((self scoreEditor))
  (om-print-document self))

(defvar *anoto-directory*  #P"/Users/jgarcia/Desktop/")

(defmethod! access-reactive-score ())

(defmethod create-anoto-file ((self scoreEditor))
  (let* ((name (gensym)))
    (print name)
    (with-open-file (out (make-pathname :directory (pathname-directory *anoto-directory*)
                                        :name (format nil "anoto-doc-~D" name)
                                        :type "txt") :direction :output  
                         :if-does-not-exist :create :if-exists :supersede) 
      (write-line (format nil "~D" name) out)
      (let* ((score (panel self))
             (fdoc (score-fdoc score))
             (size (staff-size score))
             (y (round (* size (score-top-margin score))))
             (x  (round (* size (score-left-margin score))))
            (staff (list! (staff-sys score))))
        (print ( list (fdoc (edition-values score) score size y x staff)))
        (loop for page in (page-list fdoc)
              for i = 0 then (+ i 1) do
              (write-line (format nil "(page ~D)" (+ i 1)) out)
              (loop for line in (line-list page)
                    for j = 0 then (+ j 1) do
                    (write-line-rects  score x y staff i j size out) ))))))



(defun write-line-rects (score x y staff pagenum linenum size out)
  (write-line (format nil "(multistaff ~D ~D ~D ~D)" x (+ y (* linenum (get-delta-line staff size score)))
                      (score-widht  score size) (- (get-line-size staff score size) (round (* size (score-line-space score))))) out)
    (loop for item in staff 
          (print item)
        for voicenum = 0 then (+ voicenum 1) do
        (write-a-voice score x (+ y (* linenum (get-delta-line staff size score))) item pagenum linenum size voicenum  out)))


;faire pour un poly
(defun write-a-voice (score x y staff pagenum linenum size voicenum out)
  (let ((startpos (posy (car (staff-list staff)))))
    (write-line (format nil "(staff ~D ~D ~D ~D)" x y ;a faire (+ y  (* size (nth voicenum (score-system-space score))))
                        (score-widht  score size) (get-system-size staff size)) out)
    (let* ((grap-voice (graphic-obj score))
           (fdoc (score-fdoc score))
           (elements (get-page-line-elements grap-voice fdoc pagenum linenum 0)))
      (loop for measure in elements  do
                  (anoto-write-rect measure out)))))

(defun write-a-rect (str rect out)
  (write-line (format nil "(~D ~D ~D ~D ~D)" str (first rect) (second rect)
                      (- (third rect) (first rect)) (- (fourth rect) (second rect))) out))
                         
(defmethod anoto-write-rect ((self t) out)
  (print self))

(defmethod anoto-write-rect ((self grap-group) out)
  (write-a-rect "group" (rectangle self) out)
  (loop for item in (inside self) do
        (anoto-write-rect item out)))
  
(defmethod anoto-write-rect ((self grap-measure) out)
  (write-a-rect "measure" (rectangle self) out)
  (loop for item in (inside self) do
        (anoto-write-rect item out)))

(defmethod anoto-write-rect ((self grap-chord) out)
  (write-a-rect "chord" (rectangle self) out)
  (loop for item in (inside self) do
        (anoto-write-rect item out)))

(defmethod anoto-write-rect ((self grap-note) out)
  (write-a-rect "note" (rectangle self) out))

(defmethod anoto-write-rect ((self grap-rest) out)
  (write-a-rect "rest" (rectangle self) out))


;usefull functions
;(mapcar 'class-name (get-class-precedence-list (find-class 'chord-seq-r)))

;OLD
;(defmethod export-anoto-for-object ((self chord-seq-r))
;  (print (list "exporting" self))
; (loop for chord in (chords self)
;        do (print (Lonset chord))
;           ( loop for note in (inside chord)
;                 do (export-anoto-for-object note))))

