;Lib created by Jrmie Garcia, Jean Bresson and Carlos Agon
;2011 - 2012 to create an OSC client for paper server
;this part contains the code for the boxes, main classes for osc sending and receiving

; Warning: First, seting the package where the functions will be defined 
(in-package :om)

;general OSC paramters
;server (java-osc application)
(defparameter *anoto-server* nil)
(defparameter *anoto-server-ip* "127.0.0.1")
(defparameter *anoto-port-out* 4321)
;client
(defparameter *anoto-client-ip* "127.0.0.1")
(defparameter *anoto-port-in* 1111)
(defparameter *anoto-registered* nil)
(defparameter *anoto-client-id* "OM")

;registered elements (substrates)
(defparameter *anoto-box-list* nil)
(defparameter *anoto-print* nil)



;;-----------------------------------------------------------------------------
;; We need to define a class that will contain the properties of an instance of
;; Moscou:receive box, since each individual box is associated with a background
;; process that sits and waits for OSC messages to arrive.
;;-----------------------------------------------------------------------------
;(defclass AnotoReceiveBox (ReceiveBox) 
;   ((etat     :initform nil  :initarg :etat      :accessor etat)
;    (process  :initform nil  :initarg :process   :accessor process)))

;(defmethod get-boxcallclass-fun ((self (eql 'anoto-server))) 'AnotoReceiveBox)

(om::defmethod! anoto-server ((on/off t) server-ip server-port client-id client-port)
  :initvals '(nil (eval *anoto-server-ip*) (eval *anoto-port-out*) (eval *anoto-client-id*) (eval *anoto-port-in*))
  :indoc '("on/off" "Anoto-server-ip" "Anoto-server-port" "Anoto-ID" "Anoto-portIn")
  :menuins '((0 (("ON" t) ("OFF" nil))))
  :icon 111 
  :doc  "Activate anoto functionnalities"
  (if on/off
      (progn
        (setf *anoto-server-ip* server-ip)
        (setf *anoto-port-out* server-port)
        (setf *anoto-client-id* client-id)
        (setf *anoto-port-in* client-port)
        (unless *anoto-server* (start-anoto-server))
        (anoto-send-registration)
        "anoto-server-started")
    (progn
       (anoto-send-unregistration)
       (when *anoto-server* (stop-anoto-server))
       (setf *anoto-print* nil)
        "anoto-server-exited")))


;user box to register or unregister from anoto server and specify the name (used for document management)
(defmethod! register-anoto-server ((register/unregister t))
            :initvals '(nil)
            :indoc '("register/unregister")
            :menuins '((0 (("Register" t) ("Unregister" nil))))
            :icon 128 
            :doc  "Connect to the server"
            (if register/unregister
                (anoto-send-registration)
              (anoto-send-unregistration)))

(defun anoto-send-registration()
  (anoto-osc-send  (list "/paperserver" "register"  *anoto-client-ip*  *anoto-port-in* *anoto-client-id*)))

;OLD
;(defun anoto-send-registration()
;  (om::osc-send (list "/paperserver/register" *anoto-client-ip*  *anoto-port-in* "OM") *anoto-server-ip* *anoto-port-out* )
;  (print "registration sent"))


(defun stop-anoto-server ()
  (om-stop-osc-server *anoto-server*)
  (setf *anoto-server* nil))

(defun anoto-send-unregistration()
  (anoto-osc-send (list "/paperserver" "unregister" *anoto-client-id*)))


(defun refresh-osc ( )
  (anoto-send-unregistration)
  (when *anoto-server* (stop-anoto-server))
  (setf *anoto-port-in* (+ *anoto-port-in* 1))
  (print (list "port-in" *anoto-port-in* "server-ip" *anoto-server-ip*))
  (unless *anoto-server* (start-anoto-server))
  (anoto-send-registration))

(defun anoto-osc-send (message)
  (print (list "send" message))
  (om::osc-send message *anoto-server-ip* *anoto-port-out* ))
 

;anoto server message decoder parse according to reactive-box names!
; then it is possible to get several instances of a same reactive-object!
(defun start-anoto-server ()
  (setf *anoto-server* 
        (om-start-osc-server *anoto-port-in* "localhost"  
                             #'(lambda (msg) 
                                 (let* ((message (print (om-decode-msg-or-bundle msg)))
                                        (addr-list (osc-adress-to-list (car message)))
                                        (commande (car (cdr message)))
                                        (box (find (car (last addr-list)) *anoto-box-list* :key 'name :test 'string-equal))
                                        (boxes (loop for b in *anoto-box-list* 
                                                     when (string-equal (name b) (car (last addr-list)))
                                                     collect b))
                                        )
                                   (cond 
                                    ((string-equal commande "register")
                                     (setf *anoto-print* t))
                                    ((string-equal commande "unregister")
                                     (print "unregistration received" ))
                                    ((string-equal commande "info")
                                     (print "information" )
                                     (print (cdr message)))
                                    ((string-equal commande "penstatus")
                                     (if( equal (car (cdr (cdr message))) 1 )
                                         (print "pen connected")
                                       (print "pen disconnected")))
                                    ((string-equal (car (last addr-list)) "data")
                                     )
                                    ((string-equal (car (last addr-list)) "paperserver")
                                     )
                                    ;(box
                                    ; (anoto-decode-message box commande (cdr (cdr message))))
                                    (boxes (loop for b in boxes do (anoto-decode-message b commande (cdr (cdr message)))))
                                    (t
                                     (print (format nil "Object ~A not found!" (last addr-list)))))
                                   )
                                 nil))))


;osc utils method
(defun osc-adress-to-list (str)
  (let (list)
    (loop while str do
          (let ((rep (multiple-value-list (string-until-char str "/"))))
            (setf str (second rep))
            (when (first rep) (push (first rep) list))))
    (cdr (reverse list))))

;osc method to define address namespace

(defmethod anoto-root-address ()
  (string+ "/paperserver/" *anoto-client-id* ))

(defmethod add-anoto-root-address (address)
  (string+ (anoto-root-address) "/" address))

(defmethod anoto-data-address ()
  (string+ (anoto-root-address) "/" "data"))

(defmethod add-anoto-data-address (address)
  (string+ (anoto-data-address) "/" address))


; (start-anoto-server)
; (stop-anoto-server)
